#ifndef	V_LIST_H
#define V_LIST_H

/*
 * The code below borrowed from linux/list.h
 */

#ifndef _LINUX_LIST_H
#define _LINUX_LIST_H

/*
 * Simple doubly linked list implementation.
 *
 * Some of the internal functions ("__xxx") are useful when
 * manipulating whole lists rather than single entries, as
 * sometimes we already know the next/prev entries and we can
 * generate better code by using them directly rather than
 * using the generic single-entry routines.
 */

struct list_head {
	struct list_head *next, *prev;
};

#define LIST_HEAD_INIT(name) { &(name), &(name) }

#define LIST_HEAD(name) \
	struct list_head name = LIST_HEAD_INIT(name)

#define INIT_LIST_HEAD(ptr) \
{ \
	(ptr)->next = (ptr); (ptr)->prev = (ptr); \
}

/**
 * Insert a new entry between two known consecutive entries. 
 *
 * This is only for internal list manipulation where we know
 * the prev/next entries already!
 */
#define __list_add(/*struct list_head * */ _new, \
	/* struct list_head * */ _prev, \
	/* struct list_head * */ _next) \
{ \
	(_next)->prev = (_new); \
	(_new)->next = (_next); \
	(_new)->prev = (_prev); \
	(_prev)->next = (_new); \
}

/**
 * list_add - add a new entry
 * @new: new entry to be added
 * @head: list head to add it after
 *
 * Insert a new entry after the specified head.
 * This is good for implementing stacks.
 */
#define list_add(/* struct list_head * */_new, \
	/* struct list_head * */ _head) \
{ \
	__list_add((_new), (_head), (_head)->next); \
}

/**
 * Delete a list entry by making the prev/next entries
 * point to each other.
 *
 * This is only for internal list manipulation where we know
 * the prev/next entries already!
 */
#define __list_del(/* struct list_head * */_prev, \
				  /* struct list_head * */_next) \
{ \
	(_next)->prev = (_prev); \
	(_prev)->next = (_next); \
}

/**
 * list_del - deletes entry from list.
 * @entry: the element to delete from the list.
 * Note: list_empty on entry does not return true after this,
 * the entry is in an undefined state.
 */
#define list_del(/* struct list_head * */_entry) \
{ \
	__list_del((_entry)->prev, (_entry)->next); \
}

/**
 * list_del_init - deletes entry from list and reinitialize it.
 * @entry: the element to delete from the list.
 */
#define list_del_init(/* struct list_head * */_entry) \
{ \
	__list_del((_entry)->prev, (_entry)->next); \
	INIT_LIST_HEAD(_entry); \
}

/**
 * list_empty - tests whether a list is empty
 * @head: the list to test.
 */
#define list_empty(/*struct list_head * */_head) \
	((_head)->next == (_head))

/**
 * list_splice - join two lists
 * @list: the new list to add.
 * @head: the place to add it in the first list.
 */
#define list_splice(/* struct list_head * */_list, \
	/*struct list_head * */_head) \
{ \
	struct list_head *first = (_list)->next; \
	if (first != (_list)) { \
		struct list_head *last = (_list)->prev; \
		struct list_head *at = (_head)->next; \
		first->prev = (_head); \
		(_head)->next = first; \
		last->next = at; \
		at->prev = last; \
	} \
}

/**
 * list_entry - get the struct for this entry
 * @ptr:	the &struct list_head pointer.
 * @type:	the type of the struct this is embedded in.
 * @member:	the name of the list_struct within the struct.
 */
#define list_entry(ptr, type, member) \
	((type *)((char *)(ptr)-(unsigned long)(&((type *)0)->member)))

/**
 * list_for_each	-	iterate over a list
 * @pos:	the &struct list_head to use as a loop counter.
 * @head:	the head for your list.
 */
#define list_for_each(pos, head) \
	for (pos = (head)->next /*, prefetch(pos->next)*/; \
    		pos != (head); \
        	pos = pos->next /* , prefetch(pos->next) */ )
        	
/**
 * list_for_each_safe	-	iterate over a list safe against removal of list entry
 * @pos:	the &struct list_head to use as a loop counter.
 * @n:		another &struct list_head to use as temporary storage
 * @head:	the head for your list.
 */
#define list_for_each_safe(pos, n, head) \
	for (pos = (head)->next, n = pos->next; pos != (head); \
		pos = n, n = pos->next)

#endif /* _LINUX_LIST_H */

/**
 * list_for_each_backward	-	iterate over a list in reverse dir
 * @pos:	the &struct list_head to use as a loop counter.
 * @head:	the head for your list.
 */
#define list_for_each_backward(pos, head) \
	for (pos = (head)->prev /*, prefetch(pos->prev)*/; \
    		pos != (head); \
        	pos = pos->prev /* , prefetch(pos->prev) */ )

#endif	/* V_LIST_H */
